#include <string>


#include <qtextstream.h>
#include <qstring.h>
#include <qtimer.h> 

#include <kdebug.h>


#include "mlsession.h"
#include "mlsessionhandler.h"
#include "mlhttpserver.h"

#include "syncml.h"
#include "syncmldatabase.h"

#include "xmlbinary.h"
#include "xmlparser.h"
#include "xmlbeautifier.h"
#include "syncmlresponsepacket.h"


SyncMLSession::SyncMLSession(SMLURI &  source, SMLURI &  target, MLHttpServer * communication, SMLVersion protocolVersion) :
		QObject(NULL, "SyncML Session"), doLater(), synceesDatabase(NULL), syncNode(NULL)
{
	#ifdef DEBUG_MEM
	kdDebug() << "DEBUG_MEM - SyncMLSession: created." << endl;
	#endif

	parser = NULL;
	lastErrorCode=SMLError::NoError;
	
	handler = new SMLSessionHandler(communication, protocolVersion);
	handler->setURI(source);
	handler->setRemoteURI(target);
	handler->getDevInf()->setOEM("KSyncML");
	handler->getDevInf()->setDevTyp(SMLDevInf::server);
	
	inLater=false;
	syncType = SMLAlert::None;
	syncingInProgress = false;

	//incomming signals (from the communication layer)	
	QObject::connect(communication, SIGNAL(newData(QByteArray &)),
					 this, SLOT(handleIncommingData(QByteArray &)));

	
	//outgoing signals (to the rest of the konnector)
	QObject::connect(this, SIGNAL(sessionFinishedStep()),
			handler, SLOT(slotParsingFinished()));
	
	QObject::connect(this, SIGNAL(sessionFinished()),
			handler, SLOT(slotEverythingFinished()));
	
	QObject::connect(this, SIGNAL(synceesRead(SyncMLSession::TSynceesList &, SMLURI &)),
			handler, SLOT(slotDataReceived(SyncMLSession::TSynceesList &, SMLURI &)));	

	QObject::connect(this, SIGNAL(keysRead(SyncMLSession::TKeysList &, SMLURI &)),
			handler, SLOT(slotKeysReceived(SyncMLSession::TKeysList &, SMLURI &)));	



};

SyncMLSession::~SyncMLSession()
{
	delete handler;
	
	#ifdef DEBUG_MEM
	kdDebug() << "DEBUG_MEM - SyncMLSession: destroyed." << endl;
	#endif
	
};

void SyncMLSession::handleIncommingData(QByteArray & data)
{
	delete parser;
	parser = new SyncMLParser(&data);
	QObject::connect(parser, SIGNAL(futureEvent(SMLTokenType )),
	                 this, SLOT(willReceiveEvent(SMLTokenType )));
	QObject::connect(parser, SIGNAL(newEvent(SMLTokenType, SMLTreeNode *)),
	                 this, SLOT(eventReceived(SMLTokenType, SMLTreeNode *)));
	QObject::connect(parser, SIGNAL(endParse()), this, SLOT(parserFinished()));

	parser->startParse();

	if (lastErrorCode != SMLError::NoError)
		kdDebug() << "SyncMLSession: Last error is " << lastErrorCode 
		<< " which means '" << errorToQString(lastErrorCode) << "'" << endl;
}

void SyncMLSession::parserFinished()
{
	QTimer::singleShot(0, this, SLOT(emitSessionFinishedStep()));
}

void SyncMLSession::emitSessionFinishedStep()
{
	kdDebug() << "SyncMLSession: Finished this step of the session." << endl;

	emit sessionFinishedStep();

}

void SyncMLSession::raiseError(SMLErrorType error, string description)
{
	kdDebug() << "Raise error: '" <<  errorToQString(error) << "' (#" << error << ")" << endl;
	lastErrorCode = error;
	lastErrorString = description;
}

void SyncMLSession::willReceiveEvent(SMLTokenType event)
{
	if (event == SMLToken::Sync)
	{
		if (syncingInProgress)
		{
			kdDebug() << "SyncMLSession: Continue to receive sync data." << endl;
		}
		else
		{
			kdDebug() << "SyncMLSession: Starting to receive sync data." << endl;
			syncingInProgress = true;
			synceesList.clear();
		}
	}
	
}

void SyncMLSession::emitSynceesRead()
{
	emit synceesRead(synceesList, *synceesDatabase);
	delete synceesDatabase;
	synceesDatabase = NULL;
}

void SyncMLSession::emitKeysRead()
{
	emit keysRead(keysList, *keysDatabase);
	delete keysDatabase;
	keysDatabase = NULL;
}


void SyncMLSession::eventReceived(SMLTokenType event, SMLTreeNode* currentNode)
{
	QString strEvent = eventToQString(event);
	if (!strEvent.isNull())
		kdDebug() << QString("SyncMLSession: Event ") << strEvent
		<< QString(" received. Namespace is : ") << namespaceToQString(currentNode->getType().space) << endl;


	using namespace SMLToken;
	using namespace SMLError;

	currentNode->simplifyMe();

	if (event == Final)
	{
		if(syncingInProgress)
		{
			//end of sync. normally we have everything in
			//synceesList now, so we just fire the event and reset.

			kdDebug() << "SyncMLSession: Syncing finished." << endl;
			
			syncingInProgress = false;
				
			QTimer::singleShot(0, this, SLOT(emitSynceesRead()));
				
			
			SMLTreeNode * status = handler->getResponse()->addStatus(currentMsgRef, syncNode, SMLError::Successful);
						
			SMLURI source(syncNode->getChildOfType(SMLToken::Source));
			source.addBase(handler->getRemoteURI());
			source.addDirectlyToNode(new SMLTreeNode(status,SMLToken::SourceRef), false);
		
			SMLURI target(syncNode->getChildOfType(SMLToken::Target));
			target.addBase(handler->getURI());
			target.addDirectlyToNode(new SMLTreeNode(status,SMLToken::TargetRef), false);
		}
	}
	else if (event == Sync)
	{
		syncNode = currentNode;
		kdDebug() << "SyncMLSession: Sync data finished." << endl;
		SMLTreeNode *n = currentNode->getChildOfType(SMLToken::Target);
		if (n)
		{
			synceesDatabase = new SMLURI(n);
		}
		else
		{
			handler->getResponse()->addStatus(currentMsgRef,currentNode,SMLError::BadRequest);
		}
	}
	else if (event == SyncML)
	{
		if (syncingInProgress)
		{
			//Hm. Multipacket sync.
			kdDebug() << "SyncMLSession: Multipacket sync. Waiting for more." << endl;
			SMLCmdRef  ret;
			SMLTreeNode *t = handler->getResponse()->addCommand(SMLToken::Alert,&ret);
			new SMLTreeNode(t,SMLToken::Data,QString("222"));
			//SMLTreeNode *i = 
			new SMLTreeNode(t,SMLToken::Item);
			handler->forceSend();
		}
	}
	else if (event == Map)
	{
		kdDebug() << "SyncMLSession: Map data finished." << endl;
		SMLTreeNode *n = currentNode->getChildOfType(SMLToken::Target);
		if (n)
		{
			SMLURI * key = new SMLURI(n);
			emit keysRead(keysList, *key);
			//QTimer::singleShot(0, this, SLOT(emitKeysRead()));
		}
		else
		{
			handler->getResponse()->addStatus(currentMsgRef,currentNode,SMLError::BadRequest);
		}
	}
	else if (event == MapItem)
	{
		
		//for (SMLTreeNode::nodeList::iterator
		//		i=currentNode->getChildren()->begin(); i!=currentNode->getChildren()->end(); i++ )
		//{
		SMLURI source(currentNode->getChildOfType(SMLToken::Source));
		SMLURI target(currentNode->getChildOfType(SMLToken::Target));
			
		kdDebug() << QString("SyncMLSession: Adding the key ") << endl;
			//keysList.insert("xxxxxxxx", "yyyyyyyyyyyy");
		keysList.insert(target.getQLocURI(), source.getQLocURI());
		//}
	}
	else if (event == Status) 
	{
		
		//we are only interested in status for the sync element, we don't care about SyncHrd
		
		SMLTreeNode *n = currentNode->getChildOfType(SMLToken::Cmd);
		if (n)
		{
			kdDebug() << "SyncMLSession: We got a status event for '" << n->getValue() << "' command." << endl;
			if (n->getValue() == "Sync")
			{
				kdDebug() << "SyncMLSession: Force sending of reply." << endl;
				handler->forceSend();
				
				kdDebug() << "SyncMLSession: Finishing." << endl;
				emit sessionFinished();
			}
		}
	}
	else if ((event == Add) || (event == Replace)) 
	{
		//kdDebug() << "SyncMLSession: Adding the data." << endl;
		for (SMLTreeNode::nodeList::iterator
			i=currentNode->getChildren()->begin(); i!=currentNode->getChildren()->end(); i++ )
		{
			if ((*i)->getType()==SMLToken::Item)
			{
				SMLTreeNode * dataNode = (*i)->getChildOfType(SMLToken::Data);
				
				if (dataNode)
				{
					SMLURI source((*i)->getChildOfType(SMLToken::Source));
					SMLURI target((*i)->getChildOfType(SMLToken::Target));
					SyncMLItem item;
					
					item.ID = source.getQLocURI();
					item.source = source.getQAbsoluteLocURI();
					item.target = target.getQAbsoluteLocURI();
					SMLTreeNode *n;
					n = (*i)->getParent()->getChildOfType(SMLToken::CmdID);
					if (n)
						item.cmdRef = n->getValue().toInt();
					
					item.msgRef = currentMsgRef;
					
					item.item = dataNode->getValue();
					if (event == Add)
						item.action = SyncMLItem::ADDED;
					else
						item.action = SyncMLItem::REPLACED;
					
					//kdDebug() << "SyncMLSession: Data " << "(ID = '" << item.ID << "') " << (item.action == SyncMLItem::ADDED ? "added" : "replaced")  << endl;
					
					synceesList.append(item);
				}
				else
				{
					handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::BadRequest);
				}
			}
		}
	}
	else if (event == Delete)
	{
		handler->getResponse()->addStatus(currentMsgRef,currentNode,SMLError::Successful);
	}
	else if (event==Alert)
	{
		SMLTreeNode *n = currentNode->getChildOfType(SMLToken::Data);
		SMLUInt alertType;
		if (n != NULL)
		{
			alertType = SMLStrToUInt(n->getSTLValue());
		}
		else
		{
			alertType = 0;
			raiseError(BadRequest);
		}
		
		kdDebug() << "SyncMLSession: Alert " << alertToQString(alertType) << " (no : " << alertType << ")"<< endl;
		switch (alertType) 
		{
			case SMLAlert::TwoWay:
			case SMLAlert::SlowFromClient:
			{
				if (inLater)
				{
					for (SMLTreeNode::nodeList::iterator i
					        =currentNode->getChildren()->begin(); i!=
					        currentNode->getChildren()->end(); i++)
					{
						if ((*i)->getType()==SMLToken::Item)
						{
							SMLURI source((*i)->getChildOfType(SMLToken::Source));
							source.addBase(handler->getRemoteURI());
							SMLURI target((*i)->getChildOfType(SMLToken::Target));
							target.addBase(handler->getURI());
							
							kdDebug() << "SyncMLSession: Requested database " << target.getQAbsoluteLocURI() << endl;
							if (handler->hasDatabase(target))
							{
								//SMLTreeNode *status = 
								handler->getResponse() ->
										addStatus(currentMsgRef,(*i),SMLError::RefreshRequired);
								
								
								// FIXME: Check last Sync date and send approrpriate
								handler->sendOneWaySync(target,source);
								syncType = alertType;
							}
							else
							{
								handler->getResponse() ->
										addStatus(currentMsgRef,(*i),SMLError::NotFound);
										
								raiseError(BadRequest);
							}
						}
					}
				}
				else
				{
					doLater.push(currentNode);
				}
				break;
			}
			case SMLAlert::Display:
			{
				for (SMLTreeNode::nodeList::iterator
			        	i=currentNode->getChildren()->begin(); i!=currentNode->getChildren()->end(); i++ )
				{
					if ((*i)->getType()==SMLToken::Item)
					{
						SMLTreeNode *n = (*i)->getChildOfType(SMLToken::Data);
						if (n!=NULL)
						{
							handler->recDisplay(n->getSTLValue());
							// FIXME
							handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::Successful);
							
						}
						else
						{
							handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::BadRequest);
						}
					}
				}
				currentNode->eliminateChildren();
				break;
			}
			default: //not supported alerts
			{
				handler->getResponse()->addStatus(currentMsgRef, currentNode ,SMLError::PaymentRequired);//well, why not? ;+)
				break;
			}
		}
	}
	else if (event == Get)
	{
		//string type, format;
		//SMLGetMetaTypeFormat(currentNode,&type,&format);  
		SMLType type(currentNode);
		for (SMLTreeNode::nodeList::iterator i = currentNode->getChildren()->begin(); 
							i!=  currentNode->getChildren()->end(); i++)
		{
			if ((*i)->getType() == SMLToken::Item)
			{
				SMLURI source((*i)->getChildOfType(SMLToken::Source));
				source.addBase(handler->getRemoteURI());
				SMLURI target((*i)->getChildOfType(SMLToken::Target));
				target.addBase(handler->getURI());
				
				//XXX TODO session handler should have a configurable devinf version!
				if ((target.getLocURI()=="./devinf11") || (target.getLocURI()=="./devinf10"))
				{
					if (handler->getDevInf()!=NULL)
					{
						//let's do it like multisync do.
						handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::Successful);
						
						
						kdDebug() << "SyncMLSession: sending results and DevInf inside." << endl;
						SMLTreeNode *t = handler->getResponse()->addResults(currentMsgRef,(*i));
						handler->getDevInf()->addDevInfToPacket(t);
					}
					else
					{
						kdDebug() << "SyncMLSession: oops, handler has no devinf" << endl;
						handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::NoContent);
					}
				}
				else
				{
					kdDebug() << "SyncMLSession: oops, item " << target.getLocURI().c_str() 
					<< " not found" << endl;
					handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::NotFound);
				}
			}
		}
	}
	else if (event == MsgID)
	{
		currentMsgRef = SMLStrToUInt(currentNode->getSTLValue());
	}
	else if ((event ==Put)||(event == Results))
	{
		//string type, format;
		//SMLGetMetaTypeFormat(currentNode,&type,&format);
		SMLType type(currentNode);

		for (SMLTreeNode::nodeList::iterator i
		        =currentNode->getChildren()->begin(); i!=
		        currentNode->getChildren()->end(); i++)
		{
			if ((*i)->getType()==SMLToken::Item)
			{
				SMLURI source((*i)->getChildOfType(SMLToken::Source));
				source.addBase(handler->getRemoteURI());
				SMLURI target((*i)->getChildOfType(SMLToken::Target));
				target.addBase(handler->getURI());
				//XXX FIXME check the version of the protocol and act accordingly
				if ((source.getLocURI()=="./devinf11") || (source.getLocURI()=="./devinf10"))
				{
					SMLTokenType p[] = {SMLToken::Data,SMLToken::DevInf::DevInf};
					SMLTreeNode *n;
					if ((n=(*i)->getChildOfType(2,p))!=NULL)
					{
						handler->setRemoteDevInf(new SMLDevInf(n));
						handler->recDevInf();
						handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::Successful);
					}
					else
						handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::BadRequest);
				}
				else
					handler->getResponse()->addStatus(currentMsgRef,(*i),SMLError::NotFound);
			}
		}
		currentNode->eliminateChildren();
	}
	else if (event == VerDTD)
	{
		if (currentNode->getValue() == "1.0")
		{
			handler->setVer(ONE_ZERO);
		}
		else if (currentNode->getValue() == "1.1")
		{
			handler->setVer(ONE_ONE);
		}
		else
		{
			kdWarning() << "SyncMLSession: Unsupported DTD version '" << currentNode->getValue() << "'" << endl;
			raiseError(VersionNotSupported);
		}

	}
	else if (event == VerProto)
	{
		if (currentNode->getValue() == "SyncML/1.0")
		{
			handler->setVer(ONE_ZERO);
		}
		else if (currentNode->getValue() == "SyncML/1.1")
		{
			handler->setVer(ONE_ONE);
		}
		else
		{
			kdWarning() << "SyncMLSession: Unsupported protocol version '" << currentNode->getValue() << "'" << endl;
			raiseError(VersionNotSupported);
		}
	}
	else if (event==SessionID)
	{
		handler->getResponse()->setSessionID(currentNode->getSTLValue());
	}
	else if (event==Status)
	{
		SMLCmdRef ref = {MsgRef:0xffff,CmdRef:0xffff};
		SMLUInt data=0;
		SMLTreeNode *n;
		if ((n = currentNode->getChildOfType(SMLToken::MsgRef))) 
			ref.MsgRef= SMLStrToUInt(n->getSTLValue());
		if ((n = currentNode->getChildOfType(SMLToken::CmdRef)))
			ref.MsgRef= SMLStrToUInt(n->getSTLValue());
		if ((n = currentNode->getChildOfType(SMLToken::Data)))
			data= SMLStrToUInt(n->getSTLValue());
		(*handler->getPending())[ref] = data;
	}
	else if (event==SyncBody)
	{
		if (!inLater)
		{
			inLater=true;
			while (!doLater.empty())
			{
				SMLTreeNode *n = doLater.front();
				eventReceived(n->getType(),n);
				doLater.pop();
			}
			inLater=false;
		}
	}
	else if (event==SyncHdr)
	{
		if (!currentNode->hasChildOfType(VerDTD))
			raiseError(BadRequest);

		SMLURI source(currentNode->getChildOfType(SMLToken::Source));
		handler->setRemoteURI(source);

		SMLURI target(currentNode->getChildOfType(SMLToken::Target));
		//FIXME if (target != handler->getURI()) throw error!

		//XXX TODO implement authentication
		if (currentMsgRef == 1)
			handler->getResponse()->addStatus(currentMsgRef, currentNode, AuthenticationAccepted);
		else
			handler->getResponse()->addStatus(currentMsgRef, currentNode, Successful);

	}
	else if (event==SMLToken::SyncML)
	{
	}
	else if (event==MetInf::Anchor)
	{
		//kdDebug() << "I don't know what to do." << endl;
	}

}


#include "mlsession.moc"

