// A modified copy of synchistory.h
#ifndef KSYNCML_SYNC_HISTORY_H
#define KSYNCML_SYNC_HISTORY_H

#include <qmap.h>

//#include <addressbooksyncee.h>
//#include <bookmarksyncee.h>
//#include <calendarsyncee.h>
#include <syncentry.h>

class KConfig;

namespace KSyncML {

/**
 * A small helper class to map a string (timestamp,md5sum) from and to a uid
 * and save this map on permanent storage like KConfig.
 * THE SYNTAX OF THE MAP IS AS FOLLOWS: remotekey => (localUID, remoteUID, md5sum)
 * localkey can be identical to remotekey
 */
class SyncHistoryMap
{
public:
	typedef QMap<QString, QString> Map;
	typedef struct 
	{
	  QString localUID;
	  QString remoteUID;
	  QString md5;
	} HistoryData;
	

	typedef QMap<QString, HistoryData> HistoryMap;
	
	typedef HistoryMap::Iterator Iterator;
	typedef HistoryMap::ConstIterator ConstIterator;
	
	SyncHistoryMap( const QString& fileName = QString::null );
	virtual ~SyncHistoryMap();

	void setFileName( const QString& fileName );
	QString fileName()const;

	void load();
	void save();

	bool contains( const QString& id_key ) const;
	
	QString localUID(const QString& id_key ) const;
	QString remoteUID(const QString& id_key ) const;
	QString md5(const QString& id_key ) const;
	
	
	void setLocalUID( const QString& localKey, const QString& remoteKey );
	void setRemoteUID( const QString& localKey, const QString& remoteKey );
	void setMD5( const QString& id_key, const QString& text_data );
	
	void move(const QString & old_key, const QString & new_key);
	
	
	//QString  text( const QString& id_key )const;
	//QString  localKey( const QString& id_key )const;
	
	//QString remoteKey( const QString& localKey)const;
	
	//bool contains( const QString& id_key )const;
	//bool containsKey( const QString& id_key )const;
	
	//void insert(   const QString& id_key, const QString& text_data );
	//void insertKey(   const QString& localKey, const QString& remoteKey );
	
	//void insert(   const QString& id_key, const QString& text_data, const QString& localKey );
	//void set( const SyncHistoryMap::Map& );

	SyncHistoryMap::HistoryMap map()
	{
	  return historyMap;
	}
	//SyncHistoryMap::Map keyMap()const;

	void clear();

protected:
	KConfig* config();
private:
	//Map mMap;
	//Map mKeyMap;
	HistoryMap historyMap;
	QString mFile;
	KConfig* mConf;
};



/**
	 * This is the generic base class for keeping track of additions, deletions,
	 * modifications between different syncs. In the purest version it relies on
	 * timestamps of \sa KSync::SyncEntry::timestamp for finding the changes, additions
	 * and deletions.
	 * You can change the way it operates from where it reads the information, on how the
	 * control string looks like, on howto safe it.
	 *
	 * If you want to use MD5 sum instead of a timestamp inherit from this class and implement
	 * the string method.
	 * If you want to use a different storage reimplement the load and save methods
 */
template <class Syn, class Ent>
	class SyncHistory
{
public:
	SyncHistory( const QString& file );
	virtual ~SyncHistory();

	void save ( );
	void load ( SyncHistoryMap::Map & keys );
	
	void setSyncee(Syn* syncee)
	{
	  mSyncee = syncee;
	}
	
	void processRemoteData ( SyncHistoryMap::Map & keys );
	void processRemoteKeys ( SyncHistoryMap::Map & keys );
	void processLocalData  ( SyncHistoryMap::Map & keys );
	
	//void addKeys(const SyncHistoryMap::Map & keys);
	
protected:
	virtual void      save( SyncHistoryMap* );
	virtual SyncHistoryMap*  load( const QString& );
	virtual QString string( Ent * );

private:
	SyncHistoryMap* loadAndClear();
	SyncHistoryMap* loadInternal();
	SyncHistoryMap *mMap;
	QString mFile;
	Syn *mSyncee;
};


/**
* \brief Construct a new SyncHistory
*
* Construct a new SyncHistory instance.
* @param sy The Syncee to be used
* @param file The path to the fileto either
		save the information or to read from
*
*/
template <class Sync, class Ent>
	SyncHistory<Sync, Ent>::SyncHistory(const QString& file )
	: mMap( 0l ), mFile( file )
{
}


/**
 * \brief d'tor
 */
template <class Sync, class Ent>
SyncHistory<Sync, Ent>::~SyncHistory() {
	delete mMap;
}

template <class Sync, class Ent>
	void SyncHistory<Sync, Ent>::processRemoteData ( SyncHistoryMap::Map & keys )
{
	mMap = loadInternal();
	
	bool found;
	Ent* entryNew;

	//we find the deleted records using their remoteUID
	SyncHistoryMap::HistoryMap::Iterator it;
	SyncHistoryMap::HistoryMap ma = mMap->map();
	for ( it = ma.begin(); it != ma.end(); ++it ) 
	{
		QString remoteUID = mMap->remoteUID( it.key() );
		if (remoteUID != "")
		{
			entryNew = static_cast<Ent*>( mSyncee->findEntry(remoteUID) );
		
			// if we've a UID but we can not find it in the Syncee it was removed
			if (!entryNew) 
			{
				entryNew = new Ent(mSyncee);
				QString localUID = mMap->localUID( it.key() );
				entryNew->setId(localUID);
				
				kdDebug() << "MLSyncHistory: Entry (r='" << remoteUID << "', l='" << entryNew->id() << "') is 'REMOVED' " << endl;
	
				entryNew->setState( KSync::SyncEntry::Removed );
				mSyncee->addEntry( entryNew );
			}
		}
	}

	
	
	bool shouldSave = false;
	/*
	* Now we'll search for some meta info: go through all entries
	* check if they exist if exist check if modified else it was added
	*/
	for ( entryNew = static_cast<Ent*>(mSyncee->firstEntry());
		entryNew != 0;
		entryNew = static_cast<Ent*>(mSyncee->nextEntry()) )
	{
		found = false;
	
		QString remoteUID = entryNew->id();
		QString key = keys[remoteUID];
		
		//kdDebug() << "MLSyncHistory: Remote key " << entryNew->id() << " COULD BE a key with ID " << localID << endl;
		if (mMap->contains(key))
		{
			QString localUID = mMap->localUID( key );
			
			if (localUID == "")
			{
			  localUID = remoteUID;
			}
			else if (localUID != remoteUID)
			{
				kdDebug() << "MLSyncHistory: UID " << entryNew->id() << " replaced by " << localUID << endl;
				mMap->setRemoteUID( key, remoteUID );
				shouldSave = true;
				entryNew->setId(localUID);
				keys[localUID] = key;
			}
			//else
			//	kdDebug() << "MLSyncHistory: UID " << localUID << " is remote and local" << endl;

			QString localChecksum = mMap->md5( key );
			QString remoteChecksum = string( entryNew );
			if ( localChecksum != remoteChecksum )
			{
				kdDebug() << "MLSyncHistory: Entry " << entryNew->id() << " is 'MODIFIED' " << endl;
				entryNew->setState( KSync::SyncEntry::Modified );
			}
			
		}
		else
		{
			if (entryNew->state() != KSync::SyncEntry::Removed)
			{
				kdDebug() << "MLSyncHistory: Entry with remoteUID " << entryNew->id() << " is 'ADDED' " << endl;
				entryNew->setState( KSync::SyncEntry::Added );
				
				//mMap->setLocalUID( key, remoteUID );
				mMap->setRemoteUID( key, remoteUID );
				mMap->setMD5( key, string( entryNew ) );
				shouldSave = true;
			}
		}
	}
	
	if (shouldSave)
		save( mMap );
}

template <class Sync, class Ent>
	void SyncHistory<Sync, Ent>::processLocalData(SyncHistoryMap::Map & keys)
{
	mMap = loadInternal();
	//SyncHistoryMap::Map keyMap = mMap->keyMap();
	//mMap = loadAndClear();

	kdDebug() << "MLSyncHistory: ProcessLocalData" << endl;
	
	int tempGUID = 2000;//arbitrary value
	
	for ( Ent* entry = (Ent*)mSyncee->firstEntry(); entry != 0; entry = (Ent*)mSyncee->nextEntry() ) 
	{
		/* only save meta for not deleted SyncEntries! */
		if ( entry->state() != KSync::SyncEntry::Removed )
		{
			QString key;
			if (keys.contains(entry->id()))
			{
				key = keys[entry->id()];
			}
			else
			{
				//the entry has been added on the local side.
				key = QString::number(tempGUID);
				keys[entry->id()] = key;
				tempGUID++;
			}
			
			QString localUID = entry->id();
			QString remoteUID = mMap->remoteUID( key );
			
			if ((localUID != remoteUID) && (remoteUID != ""))
			{
				kdDebug() << "MLSyncHistory: LocalUID " << localUID << " is remoteUID " << remoteUID << endl;
				entry->setId(remoteUID);
			}
			//else
				//kdDebug() << "MLSyncHistory: UID " << localUID << " is local and remote" << endl;
			
			mMap->setLocalUID( key, localUID );
			mMap->setMD5( key, string( entry ) );
		}
		else
		{
		  kdDebug() << "MLSyncHistory: Do not saving " << entry->id() << " 'coz it's removed" << endl;
		}
	}
	kdDebug() << "MLSyncHistory: ProcessLocalData end" << endl;
	save( mMap );
}


template <class Sync, class Ent>
	void SyncHistory<Sync, Ent>::processRemoteKeys ( SyncHistoryMap::Map & keys )
{
	mMap = loadInternal();
	kdDebug() << "MLSyncHistory: ProcessRemoteKeys" << endl;
	
	SyncHistoryMap::Map::ConstIterator it;
	for ( it = keys.begin(); it != keys.end(); ++it ) 
	{
		kdDebug() << "MLSyncHistory : temporary key '" 
				<< it.key() <<  "' is remote key '" 
				<< it.data() << "'" << endl;
		
		mMap->move(it.key(), it.data());
	}
	
	kdDebug() << "MLSyncHistory: ProcessRemoteKeys end" << endl;
	save( mMap );
}




/**
* @internal
* responsible for saving the SyncHistoryMap to a KConfig file
 */
template <class Sync, class Ent>
	void SyncHistory<Sync, Ent>::save( SyncHistoryMap* m) 
	{
		m->save();
	}

/**
* @internal
*
* create and load the SyncHistoryMap
 */
template <class Sync, class Ent>
SyncHistoryMap* SyncHistory<Sync, Ent>::load( const QString& f) 
{
	SyncHistoryMap *map  = new SyncHistoryMap( f );

	map->load();
	return map;
}


/**
* @internal
* Reimplement this if you for example want ot use MD5 sums.
* This implementation uses the timestamp
 */
template <class Sync, class Ent>
QString SyncHistory<Sync, Ent>::string( Ent* ent) 
{
	return ent->timestamp();
}


/**
  * @internal
 */
template <class Sync, class Ent>
		SyncHistoryMap* SyncHistory<Sync, Ent>::loadAndClear() 
{
	if(!mMap )
		mMap = load( mFile );
	
	mMap->clear();

	return mMap;
}

/**
 * @internal
 */
template <class Sync, class Ent>
SyncHistoryMap* SyncHistory<Sync, Ent>::loadInternal() 
{
	if(!mMap )
		mMap = load( mFile );

	return mMap;
}

}
#endif
