/***************************************************************************
                                 libSyncML
           Implementation of the SyncML protocol (www.syncml.org)
         Copyright (C) 2001, 2002 by Max Berger <max.berger@xslt.de>
 **************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU Lesser General Public             *
 *  License as published by the Free Software Foundation; either           *
 *  version 2.1 of the License, or (at your option) any later              *
 *  version.                                                               *
 *                                                                         *
 *  This library is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  Lesser General Public License for more details.                        *
 *                                                                         *
 *  You should have received a copy of the GNU Lesser General Public       *
 *  License along with this library; if not, write to the Free             *
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA      *
 *  02111-1307 USA                                                         *
 *                                                                         *
 ***************************************************************************/

#ifndef __SYNCML_H
#define __SYNCML_H

#define HAVE_STDINT_H 1

#include <qstring.h>

#include <string>
#include <vector>
#include <list>
#include <map>
#include <queue>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#define SMLUInt uint32_t
#define SMLErrorType SMLUInt
#define SMLAlertType SMLUInt
#define SMLTokenSpaceType uint8_t
#define SMLTokenTokenType uint8_t

#define SMLDefaultType "text/plain"
#define SMLDefaultFormat "chr"


class QTextStream;

using namespace std;


// Usually, we don't care about the namespaces, but sometimes we actually have
// to :( This gives us the opportunity to forget about them unless necessary
struct SMLTokenType {
  SMLTokenSpaceType space;
  SMLTokenTokenType token;
  //inline operator uint16_t() { return (space<<8) | token; }
  inline operator uint16_t() const { return (space<<8) | token; };
  inline bool operator ==(const SMLTokenType &a) { 
    return (a.space==space)&&(a.token==token); 
  }
  inline bool operator ==(const SMLTokenType &a) const { 
    return (a.space==space)&&(a.token==token); 
  }
  bool operator < (const SMLTokenType &a) const {
    if (a.space<space) return true;
    return a.token<token;
  }
  SMLTokenType operator = (uint16_t tvalue) {
    SMLTokenType ret;
    ret.space = (tvalue>>8); 
    ret.token = (tvalue&0xff);
    return ret;
  }
  string toDebugStr();
};

namespace SMLAlert {
  const SMLAlertType None    =   0;
  const SMLAlertType Display = 100;
  const SMLAlertType TwoWay  = 200;
  const SMLAlertType SlowFromClient  = 201;
  const SMLAlertType SlowFromServer  = 204;
  const SMLAlertType NextMessage  = 222;
}

namespace SMLError {
  const SMLErrorType NoError = 0;

  const SMLErrorType Successful = 200;
  const SMLErrorType ItemAdded = 201;
  const SMLErrorType Processing = 202;
  const SMLErrorType NonAuthorative = 203;
  const SMLErrorType NoContent = 204;
  const SMLErrorType ResetContent = 205;
  const SMLErrorType PartialContent = 206;
  const SMLErrorType ConfilctResolvedMerge = 207;
  const SMLErrorType ConflictResolvedClient = 208;
  const SMLErrorType ConflictResolvedDuplicate = 209;
  const SMLErrorType DeletedWithoutArchive = 210;
  const SMLErrorType ItemNotDeleted = 211;
  const SMLErrorType AuthenticationAccepted = 212;
  
  const SMLErrorType BadRequest = 400;
  const SMLErrorType Unauthorized = 401;
  const SMLErrorType PaymentRequired = 402;
  const SMLErrorType Forbidden = 403;
  const SMLErrorType NotFound = 404;
  const SMLErrorType CommandNotAllowed = 405;
  const SMLErrorType OptionalFeatureNotSupported = 406;
  const SMLErrorType MissingCredentials = 407;

  const SMLErrorType VersionNotSupported = 505;
  
  const SMLErrorType RefreshRequired = 508;

  const SMLErrorType LibOffset = 10000;
  const SMLErrorType Pending = LibOffset+1;
  const SMLErrorType NoDevInf = LibOffset+2;

}

namespace SMLNamespace {
  const SMLTokenSpaceType None = 0x00;       // String, Root, ...
  const SMLTokenSpaceType SyncML = 0x10;     // FD1:00
  const SMLTokenSpaceType MetInf = 0x11;     // FD1:01
  const SMLTokenSpaceType DevInf = 0x12;     // FD2:01
  const SMLTokenSpaceType OtherStart = 0x20;      // All other Namespaces
  const SMLTokenSpaceType OtherEnd = 0xf0;
}

namespace SMLToken {
  // Must be 0 sice the map returns NULL if not found...
  const SMLTokenType Illegal = {space:0,token:0x0};

  // Since this is the "main" stuff, no sub namespace
  // besides, the name for the subspace would be "SyncML" which is not good.
  const SMLTokenType Add = {space:SMLNamespace::SyncML,token:0x05};
  const SMLTokenType Alert = {space:SMLNamespace::SyncML,token:0x06};
  const SMLTokenType Archive = {space:SMLNamespace::SyncML,token:0x07};
  const SMLTokenType Atomic = {space:SMLNamespace::SyncML,token:0x08};
  const SMLTokenType Chal = {space:SMLNamespace::SyncML,token:0x09};
  const SMLTokenType Cmd = {space:SMLNamespace::SyncML,token:0x0a};
  const SMLTokenType CmdID = {space:SMLNamespace::SyncML,token:0x0b};
  const SMLTokenType CmdRef = {space:SMLNamespace::SyncML,token:0x0c};
  const SMLTokenType Copy = {space:SMLNamespace::SyncML,token:0x0d};
  const SMLTokenType Cred = {space:SMLNamespace::SyncML,token:0x0e};
  const SMLTokenType Data = {space:SMLNamespace::SyncML,token:0x0f};
  const SMLTokenType Delete = {space:SMLNamespace::SyncML,token:0x10};
  const SMLTokenType Exec = {space:SMLNamespace::SyncML,token:0x11};
  const SMLTokenType Final = {space:SMLNamespace::SyncML,token:0x12};
  const SMLTokenType Get = {space:SMLNamespace::SyncML,token:0x13};
  const SMLTokenType Item = {space:SMLNamespace::SyncML,token:0x14};
  const SMLTokenType Lang = {space:SMLNamespace::SyncML,token:0x15};
  const SMLTokenType LocName = {space:SMLNamespace::SyncML,token:0x16};
  const SMLTokenType LocURI = {space:SMLNamespace::SyncML,token:0x17};
  const SMLTokenType Map = {space:SMLNamespace::SyncML,token:0x18};
  const SMLTokenType MapItem = {space:SMLNamespace::SyncML,token:0x19};
  const SMLTokenType Meta = {space:SMLNamespace::SyncML,token:0x1a};
  const SMLTokenType MsgID = {space:SMLNamespace::SyncML,token:0x1b};
  const SMLTokenType MsgRef = {space:SMLNamespace::SyncML,token:0x1c};
  const SMLTokenType NoResp = {space:SMLNamespace::SyncML,token:0x1d};
  const SMLTokenType NoResults = {space:SMLNamespace::SyncML,token:0x1e};
  const SMLTokenType Put = {space:SMLNamespace::SyncML,token:0x1f};
  const SMLTokenType Replace = {space:SMLNamespace::SyncML,token:0x20};
  const SMLTokenType RespURI = {space:SMLNamespace::SyncML,token:0x21};
  const SMLTokenType Results = {space:SMLNamespace::SyncML,token:0x22};
  const SMLTokenType Search = {space:SMLNamespace::SyncML,token:0x23};
  const SMLTokenType Sequence = {space:SMLNamespace::SyncML,token:0x24};
  const SMLTokenType SessionID = {space:SMLNamespace::SyncML,token:0x25};
  const SMLTokenType SftDel = {space:SMLNamespace::SyncML,token:0x26};
  const SMLTokenType Source = {space:SMLNamespace::SyncML,token:0x27};
  const SMLTokenType SourceRef = {space:SMLNamespace::SyncML,token:0x28};
  const SMLTokenType Status = {space:SMLNamespace::SyncML,token:0x29};
  const SMLTokenType Sync = {space:SMLNamespace::SyncML,token:0x2a};
  const SMLTokenType SyncBody = {space:SMLNamespace::SyncML,token:0x2b};
  const SMLTokenType SyncHdr = {space:SMLNamespace::SyncML,token:0x2c};
  const SMLTokenType SyncML = {space:SMLNamespace::SyncML,token:0x2d};
  const SMLTokenType Target = {space:SMLNamespace::SyncML,token:0x2e};
  const SMLTokenType TargetRef = {space:SMLNamespace::SyncML,token:0x2f};
  const SMLTokenType Reserved = {space:SMLNamespace::SyncML,token:0x30};
  const SMLTokenType VerDTD = {space:SMLNamespace::SyncML,token:0x31};
  const SMLTokenType VerProto = {space:SMLNamespace::SyncML,token:0x32};
  
  namespace MetInf {
    const SMLTokenType Anchor = {space:SMLNamespace::MetInf,token:0x05};
    const SMLTokenType EMI = {space:SMLNamespace::MetInf,token:0x06};
    const SMLTokenType Format = {space:SMLNamespace::MetInf,token:0x07};
    const SMLTokenType FreeID = {space:SMLNamespace::MetInf,token:0x08};
    const SMLTokenType FreeMem = {space:SMLNamespace::MetInf,token:0x09};
    const SMLTokenType Last = {space:SMLNamespace::MetInf,token:0x0a};
    const SMLTokenType Mark = {space:SMLNamespace::MetInf,token:0x0b};
    const SMLTokenType MaxMsgSize = {space:SMLNamespace::MetInf,token:0x0c};
    const SMLTokenType Mem = {space:SMLNamespace::MetInf,token:0x0d};
    const SMLTokenType MetInf = {space:SMLNamespace::MetInf,token:0x0e};
    const SMLTokenType Next = {space:SMLNamespace::MetInf,token:0x0f};
    const SMLTokenType NextOnce = {space:SMLNamespace::MetInf,token:0x10};
    const SMLTokenType SharedMem = {space:SMLNamespace::MetInf,token:0x11};
    const SMLTokenType Size = {space:SMLNamespace::MetInf,token:0x12};
    const SMLTokenType Type = {space:SMLNamespace::MetInf,token:0x13};
    const SMLTokenType Version = {space:SMLNamespace::MetInf,token:0x14};
  } 

  namespace DevInf {
    const SMLTokenType CTCap = {space:SMLNamespace::DevInf,token:0x05}; 
    const SMLTokenType CTType = {space:SMLNamespace::DevInf,token:0x06};
    const SMLTokenType DataStore = {space:SMLNamespace::DevInf,token:0x07};
    const SMLTokenType DataType = {space:SMLNamespace::DevInf,token:0x08};
    const SMLTokenType DevId = {space:SMLNamespace::DevInf,token:0x09};
    const SMLTokenType DevInf = {space:SMLNamespace::DevInf,token:0x0a};
    const SMLTokenType DevTyp = {space:SMLNamespace::DevInf,token:0x0b};
    const SMLTokenType DisplayName = {space:SMLNamespace::DevInf,token:0x0c};
    const SMLTokenType DSMem = {space:SMLNamespace::DevInf,token:0x0d};
    const SMLTokenType Ext = {space:SMLNamespace::DevInf,token:0x0e};
    const SMLTokenType FwV = {space:SMLNamespace::DevInf,token:0x0f};
    const SMLTokenType HwV = {space:SMLNamespace::DevInf,token:0x10};
    const SMLTokenType Man = {space:SMLNamespace::DevInf,token:0x11};
    const SMLTokenType MaxGUIDSize = {space:SMLNamespace::DevInf,token:0x12};
    const SMLTokenType MaxID = {space:SMLNamespace::DevInf,token:0x13};
    const SMLTokenType MaxMem = {space:SMLNamespace::DevInf,token:0x14};
    const SMLTokenType Mod = {space:SMLNamespace::DevInf,token:0x15};
    const SMLTokenType OEM = {space:SMLNamespace::DevInf,token:0X16};
    const SMLTokenType ParamName = {space:SMLNamespace::DevInf,token:0x17};
    const SMLTokenType PropName = {space:SMLNamespace::DevInf,token:0x18};
    const SMLTokenType Rx = {space:SMLNamespace::DevInf,token:0x19};
    const SMLTokenType RxPref = {space:SMLNamespace::DevInf,token:0x1a};
    const SMLTokenType SharedMem = {space:SMLNamespace::DevInf,token:0x1b};
    const SMLTokenType Size = {space:SMLNamespace::DevInf,token:0x1c};
    const SMLTokenType SourceRef = {space:SMLNamespace::DevInf,token:0x1d};
    const SMLTokenType SwV = {space:SMLNamespace::DevInf,token:0x1e};
    const SMLTokenType SyncCap = {space:SMLNamespace::DevInf,token:0x1f};
    const SMLTokenType SyncType = {space:SMLNamespace::DevInf,token:0x20};
    const SMLTokenType Tx = {space:SMLNamespace::DevInf,token:0x21};
    const SMLTokenType TxPref = {space:SMLNamespace::DevInf,token:0x22};
    const SMLTokenType ValEnum = {space:SMLNamespace::DevInf,token:0x23};
    const SMLTokenType VerCT = {space:SMLNamespace::DevInf,token:0x24};
    const SMLTokenType VerDTD = {space:SMLNamespace::DevInf,token:0x25};
    const SMLTokenType XNam = {space:SMLNamespace::DevInf,token:0x26};
    const SMLTokenType XVal = {space:SMLNamespace::DevInf,token:0x27};
   }

  // This is the magic not-really existing root token. Please don't confuse it
  // with the "real" root token <SyncML> 
  const SMLTokenType Root = {space:SMLNamespace::None,token:0xf0}; 
  const SMLTokenType String = {space:SMLNamespace::None,token:0xf1};
  const SMLTokenTokenType AlienToken = 0xf2;
}

namespace SMLSyncType {
  const string twoway = "1";
  const string slow = "2";
  const string onewayFromClient = "3";
  const string refreshFromClient = "4";
  const string onewayFromServer = "5";
  const string refreshFromServer = "6";
  const string serverAlerted = "7";
}


namespace SMLMimeType {
  const QString devinfxml = "application/vnd.syncml-devinf+xml";
}


typedef enum { UNKNOWN, ONE_ZERO, ONE_ONE } SMLVersion;



class SMLNamespaceContainer;
class SMLFlattener;
class SMLTreeNode;
class SMLResponsePacket;
class SyncMLParser;
class SyncMLDatabase;
  
struct SMLCmdRef{
  SMLUInt MsgRef;
  SMLUInt CmdRef;
};

inline bool operator< (const SMLCmdRef& lhs, const SMLCmdRef& rhs) {
  if (lhs.MsgRef == rhs.MsgRef) return lhs.CmdRef < rhs.CmdRef;
  return lhs.MsgRef < rhs.MsgRef;
}

typedef map<SMLCmdRef,SMLErrorType> SMLCmdMapType;

class SMLType {
 private:
  string format;
  string type;
 public:
  SMLType(string typ=SMLDefaultType,string form=SMLDefaultFormat) {
    type=typ;
    format=form;
    };
  SMLType(SMLTreeNode *parent);
  string getType() { return type; }
  string getFormat() { return format; }
  void addToNode(SMLTreeNode *parent);
};


QString SMLTimestamp();
string SMLCreateUID();
QString SMLUIntToStr(SMLUInt x);

SMLUInt SMLStrToUInt(string s);


//DEBUG FUNCTIONS
QString eventToQString(SMLTokenType event);
QString namespaceToQString(SMLTokenSpaceType space);
QString alertToQString(SMLAlertType token);
QString errorToQString(SMLAlertType token);


#endif
