/***************************************************************************
                                 libSyncML
           Implementation of the SyncML protocol (www.syncml.org)
         Copyright (C) 2001, 2002 by Max Berger <max.berger@xslt.de>
 **************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This library is free software; you can redistribute it and/or          *
 *  modify it under the terms of the GNU Lesser General Public             *
 *  License as published by the Free Software Foundation; either           *
 *  version 2.1 of the License, or (at your option) any later              *
 *  version.                                                               *
 *                                                                         *
 *  This library is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      *
 *  Lesser General Public License for more details.                        *
 *                                                                         *
 *  You should have received a copy of the GNU Lesser General Public       *
 *  License along with this library; if not, write to the Free             *
 *  Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA      *
 *  02111-1307 USA                                                         *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <string>
#include <iomanip>

#include <qstring.h>
#include <qdatetime.h>

#include "syncml.h"
#include "xmlbinary.h"
#include "xmlparser.h"
#include "xmlbeautifier.h"
#include "syncmlresponsepacket.h"
#include "mlsession.h"
#include "mlsessionhandler.h"

QString eventToQString(SMLTokenType event)
{
	using namespace SMLToken;
	using namespace SMLError;

	if (event==Final)
		return "Final";
	if (event==Alert)
		return "Alert";
	else if (event ==Get)
		return("Get");
	else if (event == MsgID)
		return("MsgID");
	else if ((event ==Put)||(event == Results))
		return("Put or Results");
	else if (event==VerDTD)
		return("VerDTD");
	else if (event==VerProto)
		return("VerProto");
	else if (event==SessionID)
		return("SessionID");
	else if (event==Status)
		return("Status");
	else if (event==SyncBody)
		return("SyncBody");
	else if (event==SyncBody)
		return("Sync");
	else if (event==SyncHdr)
		return "SyncHdr";
	else if (event==SMLToken::SyncML)
		return "SMLToken:SyncML";
	else if (event==MetInf::Anchor)
		return "MetaInf:Anchor";
	else if (event==Add)
		return "Add";
	else
		return QString();//null string

}

QString namespaceToQString(SMLTokenSpaceType space)
{
	using namespace SMLNamespace;
	if (space==MetInf)
		return "MetInf";
	else if (space == None)
		return("None");
	else if (space == SyncML)
		return("SyncML");
	else if (space == DevInf)
		return("DevInf");
	else if (space==OtherStart)
		return("OtherStart");
	else if (space==OtherEnd)
		return("OtherEnd");
	else
		return "<Unknown Namespace>";

}
QString alertToQString(SMLAlertType token)
{
	using namespace SMLAlert;

	if (token==None)
		return "None";
	else if (token == TwoWay)
		return("TwoWay");
	else if (token == SlowFromClient)
		return("SlowFromClient");
	else if (token == SMLAlert::Display)
		return("Display");
	else if (token == SMLAlert::NextMessage)
		return("NextMessage");
	else
		return "<Unknown Alert>";

}

QString errorToQString(SMLAlertType token)
{
	using namespace SMLError;

	if      (token == NoError)
		return "NoError";
	else if (token == Successful)
		return("Successful");
	else if (token == ItemAdded)
		return("ItemAdded");
	else if (token == Processing)
		return("Processing");
	else if (token == NonAuthorative)
		return("NonAuthorative");
	else if (token == NoContent)
		return("NoContent");
	else if (token == ResetContent)
		return("ResetContent");
	else if (token == PartialContent)
		return("PartialContent");
	else if (token == ConfilctResolvedMerge)
		return("ConfilctResolvedMerge");
	else if (token == ConflictResolvedClient)
		return("ConflictResolvedClient");
	else if (token == ConflictResolvedDuplicate)
		return("ConflictResolvedDuplicate");
	else if (token == DeletedWithoutArchive)
		return("DeletedWithoutArchive");
	else if (token == ItemNotDeleted)
		return("ItemNotDeleted");
	else if (token == AuthenticationAccepted)
		return("AuthenticationAccepted");
	else if (token == BadRequest)
		return("BadRequest");
	else if (token == Unauthorized)
		return("Unauthorized");
	else if (token == PaymentRequired)
		return("PaymentRequired");
	else if (token == Forbidden)
		return("Forbidden");
	else if (token == NotFound)
		return("NotFound");
	else if (token == CommandNotAllowed)
		return("CommandNotAllowed");
	else if (token == OptionalFeatureNotSupported)
		return("OptionalFeatureNotSupported");
	else if (token == MissingCredentials)
		return("MissingCredentials");
	else if (token == VersionNotSupported)
		return("VersionNotSupported");
	else if (token == LibOffset)
		return("LibOffset");
	else if (token == Pending)
		return("Pending");
	else if (token == NoDevInf)
		return("NoDevInf");
	else
		return "<Unknown Error>";
}




QString SMLTimestamp() 
{
  //20040304T192746Z
  QDateTime time = QDateTime::currentDateTime();
  QString res;
  QString tmp = time.date().toString("yyyyMMdd");
  res.append(tmp);
  res.append("T");
  
  tmp = time.time().toString("hhmmss");
  res.append(tmp);
  res.append("Z");
  return res;
}

string SMLCreateUID() {
//XXX TODO generation of timestamps  
  //strstream s;
  //s << SMLTimestamp() << "LIBSYNCML" << setw(5) << setfill('0') << rand();
  //return string(s.str(),s.pcount());
  return string("AAAALIBSYNCML");
}

QString SMLUIntToStr(SMLUInt x) {
  //strstream s;
  //s << x;
  //return string(s.str(),s.pcount());
  QString s;
  s.setNum(x);
  return s;//string(s.ascii());
}

SMLUInt SMLStrToUInt(string s) {
  // FIXME - this is limited to long types!
   return atol(s.c_str());
}


SMLType::SMLType(SMLTreeNode *parent) {
  type=SMLDefaultType;
  format=SMLDefaultFormat;
  SMLTreeNode *m = parent->getChildOfType(SMLToken::Meta);
  if (m!=NULL) {
    SMLTreeNode* n;
    if ((n = m->getChildOfType(SMLToken::MetInf::Type))!=NULL)
      type = n->getSTLValue();
    if ((n = m->getChildOfType(SMLToken::MetInf::Format))!=NULL)
      format = n->getSTLValue();
  }
}

void SMLType::addToNode(SMLTreeNode *parent) {
  if ((format!=SMLDefaultFormat)||(type!=SMLDefaultType)) {
    SMLTreeNode *m = new SMLTreeNode(parent,SMLToken::Meta);
    if (type!=SMLDefaultType) 
      new SMLTreeNode(m,SMLToken::MetInf::Type, QString(type.c_str()));
    if (format!=SMLDefaultFormat)
	new SMLTreeNode(m,SMLToken::MetInf::Format,QString(format.c_str()));
  }
}

